/*
** Module   : LockTableUpdater.java
** Abstract : Manages metadata lock records for a particular, primary database
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20131101 Created initial version.
** 002 ECF 20140116 Use session ID as user number (id) for lock events.
*/

package com.goldencode.p2j.persist.meta;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.*;
import com.goldencode.proxy.*;

/**
 * A {@link com.goldencode.p2j.persist.lock.LockListener lock listener implementation} which
 * receives record lock events and updates the metadata lock table accordingly. This involves the
 * creation, lookup, and deletion of metadata lock records. To avoid impacting the performance of
 * the current session's database operations, most of this work is done within a separate,
 * dedicated, server thread.
 */
public final class LockTableUpdater
implements LockListener
{
   /** Logger */
   private static final Logger log = LogHelper.getLogger(LockTableUpdater.class);
   
   /** Worker which processes record lock events in a dedicated server thread */
   private final UpdateWorker worker;
   
   /** Numeric ID of the user in the current session */
   private final ContextLocal<integer> sessionID = new ContextLocal<integer>()
   {
      /** Security manager */
      private final SecurityManager securityManager;
      
      /** List of reclaimed user numbers */
      private final List<integer> reclaimedSessionIDs;
      
      /** Highest allocated session ID to date (may no longer be active) */
      private int highSessionID = 0;
      
      // Constructor for this anonymous inner class
      {
         securityManager = SecurityManager.getInstance();
         reclaimedSessionIDs = (securityManager == null ? new LinkedList<integer>() : null);
      }
      
      @Override
      protected integer initialValue()
      {
         integer sessionID;
         
         if (securityManager != null)
         {
            sessionID = new integer(securityManager.getSessionId());
         }
         else
         {
            // fallback method, though we should always have a security manager in server mode
            synchronized (this)
            {
               if (reclaimedSessionIDs.size() > 0)
               {
                  sessionID = reclaimedSessionIDs.remove(0);
               }
               else
               {
                  sessionID = new integer(++highSessionID);
               }
            }
         }
         
         return sessionID;
      }
      
      protected void cleanup(integer value)
      {
         if (securityManager == null)
         {
            synchronized (this)
            {
               reclaimedSessionIDs.add(value);
            }
         }
      }
   };
   
   /**
    * Constructor.
    * 
    * @param   metaDatabase
    *          Metadata database.
    */
   public LockTableUpdater(Database metaDatabase)
   {
      this.worker = new UpdateWorker(PersistenceFactory.getInstance(metaDatabase));
   }
   
   /**
    * This method is called within a user context to force any pending lock events to be
    * processed. It guarantees that any events that are pending in the queue at the time this
    * method is invoked will have been processed by the time this method returns. However,
    * other events may be added by other user contexts in the meantime, and these are not
    * guaranteed to have been processed.
    */
   public void flushQueue()
   {
      worker.forceFlush();
   }
   
   /**
    * Stop processing record lock events for the associated, primary database. This will terminate
    * the update worker and its processing thread.
    */
   public void terminate()
   {
      worker.terminate();
   }
   
   /**
    * Method which is called by a {@link LockManager} in response to a lock event. Does some
    * preliminary processing and sanity checking, then queues the event for further processing on
    * a dedicated thread.
    * 
    * @param   event
    *          Object which describes the lock event.
    */
   public void lockChanged(RecordLockEvent event)
   {
      LockType oldType = event.getOldType();
      LockType newType = event.getNewType();
      
      // if old and new lock types are both NONE, nothing to do, but log a warning, because this
      // suggests a problem
      if (LockType.NONE.equals(oldType) && LockType.NONE.equals(newType))
      {
         if (log.isLoggable(Level.WARNING))
         {
            log.log(Level.WARNING, "Unexpected lock event:  " + event);
            
            return;
         }
      }
      
      event.setUserid(sessionID.get());
      
      worker.addEvent(event);
   }
   
   /**
    * A worker containing a queue to which record lock events are written from multiple user
    * threads/contexts, and from which those events are read in a dedicated, daemon thread. The
    * idea is to offload the potentially more time-consuming work of querying and updating the
    * metadata database from a dedicated thread, to minimize any performance penalty of this work
    * on user sessions. Since this naturally introduces the possibility of the lock metadata to
    * get out of sync with the actual lock states in the lock manager, a synchronization mechanism
    * is used to ensure any pending updates are flushed to the database before any query against
    * this table is executed. This way, only those user sessions actually querying the lock
    * metadata are affected by any lag in updating the metadata from record lock events.
    * <p>
    * Note that some analysis of the lock event has to occur in the user session in which the
    * lock event occurred, but as much of the work as possible is deferred until the separate
    * thread processes the events.
    */
   private static class UpdateWorker
   implements Runnable
   {
      /** Metadata lock table name */
      private static final String LOCK_TABLE = "meta_lock";
      
      /** HQL query to find a lock record by lock recid and user */
      private static final String HQL_LOCK =
         "select m.id from MetaLockImpl m where m.lockRecid = ? and m.lockUsr = ?";
      
      /** Parameter types for lock record query */
      private static final Type[] TYPES_LOCK;
      
      /** HQL query to find a file number by file name */
      private static final String HQL_FILENUM =
         "select m.fileNumber from MetaFileImpl m where m.__fileName = ?";
      
      /** Parameter types for file number query */
      private static final Type[] TYPES_FILENUM;
      
      /** MetaLockImpl DMO implementation class */
      private static final Class<?> dmoClass;
      
      /** Map of <code>MinimalLock</code> methods to <code>MetaLockImpl</code> methods */
      private static final Map<Method, Method> methodMap = new HashMap<Method, Method>();
      
      static
      {
         String clsName = String.format("%s._meta.impl.MetaLockImpl", DMOIndex.getBasePackage());
         try
         {
            dmoClass = Class.forName(clsName);
            
            // map MinimalLock methods to corresponding dmoClass methods; if we try to use
            // MinimalLock methods with a MetaLockImpl object, we will raise
            // IllegalArgumentException during Method.invoke(), because our object will not be an
            // instance of MinimalLock
            for (Method ifcMeth : MinimalLock.class.getMethods())
            {
               Method dmoMeth = dmoClass.getMethod(ifcMeth.getName(),
                                                   ifcMeth.getParameterTypes());
               methodMap.put(ifcMeth, dmoMeth);
            }
         }
         catch (ClassNotFoundException | NoSuchMethodException exc)
         {
            throw new RuntimeException("Error preparing MetaLock implementation class", exc);
         }
         
         TYPES_LOCK = DBUtils.makeTypeArray(new Object[] { new recid(), new integer() });
         TYPES_FILENUM = DBUtils.makeTypeArray(new Object[] { new character() });
      }
      
      /** Event queue to which many threads can write and from which only one thread reads */
      private final List<RecordLockEvent> eventQueue = new LinkedList<RecordLockEvent>();
      
      /** Persistence helper object for metadata database */
      private final Persistence persistence;
      
      /** Proxy object for the current metadata lock DMO */
      private final MinimalLock proxy;
      
      /** Invocation handler which maps proxy method calls to underlying DMO methods */
      private final Handler handler = new Handler();
      
      /** Update worker thread */
      private final AssociatedThread workThread;
      
      /** File number cache */
      private final Map<character, integer> fileNums = new HashMap<character, integer>();
      
      /** Termination flag */
      private boolean done = false;
      
      /** Lock needed to read from queue; must be obtained first if write lock is needed also */
      private Object readLock = new Object();
      
      /** Lock needed to write to queue; must be obtained second if read lock is needed also */
      private Object writeLock = new Object();
      
      /** Unique and sequential ID of the last event added to the queue */
      private long lastAddedID = -1L;
      
      /** Unique and sequential ID of the last event read from the queue and processed */
      private long lastProcessedID = -1L;
      
      /**
       * Constructor.
       * 
       * @param   persistence
       *          Persistence helper object for metadata database.
       */
      UpdateWorker(Persistence persistence)
      {
         super();
         
         this.persistence = persistence;
         this.proxy = ProxyFactory.getProxy(dmoClass,
                                            new Class<?>[] { MinimalLock.class },
                                            handler);
         
         workThread = new AssociatedThread(this);
         workThread.setDaemon(true);
         workThread.start();
      }
      
      /**
       * This method executes in the worker thread, reading record lock events from the queue when
       * they become available, and updating the metadata database accordingly.
       */
      @Override
      public void run()
      {
         while (true)
         {
            // check if we should terminate event processing
            synchronized (writeLock)
            {
               if (done)
               {
                  break;
               }
            }
            
            // process any pending events
            processPendingEvents();
            
            // wait for new events to be added to the queue if it is still empty
            synchronized (writeLock)
            {
               while (!done && eventQueue.isEmpty())
               {
                  try
                  {
                     writeLock.wait();
                  }
                  catch (InterruptedException exc)
                  {
                     // may be spurious, ignore;  if interrupted due to termination, we will
                     // detect this in the termination check above
                  }
               }
            }
         }
      }
      
      /**
       * Add a record lock event to the queue, locking out other write and read operations until
       * the queue is updated. Notify a the worker thread when the new event has been added.
       * 
       * @param   event
       *          Record lock event to be added.
       */
      void addEvent(RecordLockEvent event)
      {
         synchronized (writeLock)
         {
            eventQueue.add(event);
            lastAddedID = event.getEventID();
            writeLock.notify();
         }
      }
      
      /**
       * This method is called within a user context to force any pending lock events to be
       * processed. It guarantees that any events that are pending in the queue at the time this
       * method is invoked will have been processed by the time this method returns. However,
       * other events may be added by other user contexts in the meantime, and these are not
       * guaranteed to have been processed.
       */
      void forceFlush()
      {
         long markerID = -1L;
         long lastProcessed = -1L;
         
         synchronized (writeLock)
         {
            markerID = lastAddedID;
            lastProcessed = lastProcessedID;
         }
         
         synchronized (readLock)
         {
            while (lastProcessed < markerID)
            {
               try
               {
                  readLock.wait();
               }
               catch (InterruptedException exc)
               {
                  if (done)
                  {
                     break;
                  }
               }
               
               synchronized (writeLock)
               {
                  lastProcessed = lastProcessedID;
               }
            }
         }
      }
      
      /**
       * Set a flag to terminate the queue, removing any pending events which have not yet been
       * processed. The worker thread is interrupted.
       * <p>
       * This method should invoked when the primary database associated with the enclosing lock
       * table updater is deregistered, as part of cleaning up the associated metadata database.
       */
      void terminate()
      {
         synchronized (writeLock)
         {
            done = true;
            eventQueue.clear();
            workThread.interrupt();
         }
      }
      
      /**
       * Process all pending record lock events in the queue. Events are read and removed from the
       * queue synchronously, such that no new events may be added during the read/remove
       * operation. However, they are then processed asynchronously, so new events may be added
       * while lock metadata is being updated.
       */
      private void processPendingEvents()
      {
         RecordLockEvent[] pending = null;
         
         synchronized (readLock)
         {
            // collect all pending events (if any)
            synchronized (writeLock)
            {
               int size = eventQueue.size();
               if (size > 0)
               {
                  pending = new RecordLockEvent[size];
                  Iterator<RecordLockEvent> iter = eventQueue.iterator();
                  for (int i = 0; iter.hasNext(); i++)
                  {
                     pending[i] = iter.next();
                     iter.remove();
                  }
               }
            }
            
            // process pending events (if any); note that new events can be added to the queue
            // while we are doing this
            if (pending != null)
            {
               int len = pending.length;
               for (int i = 0; i < len; i++)
               {
                  processEvent(pending[i]);
               }
               
               synchronized (writeLock)
               {
                  lastProcessedID = pending[len - 1].getEventID();
               }
            }
            
            // give any threads waiting in forceFlush a chance to proceed
            readLock.notifyAll();
         }
      }
      
      /**
       * Process an individual record lock event, updating the metadata lock table as appropriate.
       * If the lock event represents a newly acquired lock, create a new lock record, otherwise
       * find the existing lock record associated with the event, which must represent a change of
       * the state of an existing lock. If that change of state represents a lock release, delete
       * the associated lock record from the lock table.
       * 
       * @param   event
       *          Record lock event describing a new lock or a change in state of an existing
       *          lock.
       */
      private void processEvent(RecordLockEvent event)
      {
         Persistable dmo = null;
         
         try
         {
            boolean commit = persistence.beginTransaction();
            
            LockType oldType = event.getOldType();
            LockType newType = event.getNewType();
            
            if (LockType.NONE.equals(oldType))
            {
               dmo = createRecord(event);
            }
            else
            {
               dmo = findRecord(event);
            }
            
            if (LockType.NONE.equals(newType))
            {
               persistence.delete(dmo);
            }
            else
            {
               String code = (newType.isExclusive() ? "X" : "S"); // TODO:  get this right
               proxy.setLockFlags(new character(code));
               boolean isTable = (event.getIdentifier() instanceof TableIdentifier);
               proxy.setLockType(new character(isTable ? "TAB" : "REC"));
               persistence.save(dmo, dmo.getId());
            }
            
            persistence.flush();
            
            if (commit)
            {
               persistence.commit();
            }
            
            if (log.isLoggable(Level.FINEST))
            {
               log.log(Level.FINEST, event.toString());
            }
         }
         catch (Exception exc)
         {
            try
            {
               persistence.rollback();
            }
            catch (PersistenceException pe)
            {
               // error already logged in finally block
            }
            finally
            {
               if (log.isLoggable(Level.SEVERE))
               {
                  String msg = String.format("Error updating _lock table for %s (%s)",
                                             persistence.getDatabase().toString(),
                                             event.toString());
                  log.log(Level.SEVERE, msg, exc);
               }
            }
         }
         finally
         {
            if (dmo != null)
            {
               RecordIdentifier metaIdent = new RecordIdentifier(dmoClass.getName(), dmo.getId());
               try
               {
                  persistence.lock(LockType.NONE, metaIdent, true);
               }
               catch (LockUnavailableException exc)
               {
                  // won't be thrown from lock release
               }
            }
         }
      }
      
      /**
       * Extract the locked record's positive, integral primary key from the record lock event, or
       * <code>0</code> if the event references a table lock.
       * 
       * @param   event
       *          A record lock event.
       * 
       * @return  See above.
       */
      private long getLockRecordID(RecordLockEvent event)
      {
         Serializable recordID = event.getIdentifier().getRecordID();
         long recID = (recordID instanceof Long ? (Long) recordID : 0L);
         
         return recID;
      }
      
      /**
       * Find the metadata lock record associated with the user ID and lock record ID specified by
       * the given record lock event.
       * 
       * @param   event
       *          Record lock event.
       * 
       * @return  Metadata lock record, if found.
       * 
       * @throws  PersistenceException
       *          if a lock record cannot be found for the given event, or if there is an error
       *          executing the query.
       */
      private Persistable findRecord(RecordLockEvent event)
      throws PersistenceException
      {
         recid lockRec = new recid(getLockRecordID(event));
         integer lockUsr = event.getUserid();
         Object[] args = new Object[] { lockRec, lockUsr };
         List<Long> list = persistence.list(HQL_LOCK, args, TYPES_LOCK, 1, 0, true);
         Long id = (list != null ? list.get(0) : null);
         Persistable dmo = null;
         
         if (id != null)
         {
            RecordIdentifier metaIdent = new RecordIdentifier(dmoClass.getName(), id);
            persistence.lock(LockType.EXCLUSIVE, metaIdent, true);
            dmo = persistence.quickLoad(metaIdent);
         }
         
         if (dmo == null)
         {
            throw new PersistenceException("No lock record found for lock event:  " + event);
         }
         
         handler.setDelegate(dmo);
         
         return dmo;
      }
      
      /**
       * Create a metadata lock record for the given record lock event.
       * 
       * @param   event
       *          Event which specifies information to be stored in the lock record.
       * 
       * @return  New lock record.
       * 
       * @throws  PersistenceException
       *          if there is any error querying related metadata, or if there is a reflection
       *          error instantiating the DMO.
       */
      private Persistable createRecord(RecordLockEvent event)
      throws PersistenceException
      {
         Persistable dmo = null;
         
         try
         {
            // create new DMO instance and back the proxy with it
            dmo = (Persistable) dmoClass.newInstance();
            handler.setDelegate(dmo);
            
            // assign primary key
            Long key = (Long) persistence.nextPrimaryKey(LOCK_TABLE);
            dmo.setId(key);
            
            // look up _file._file-number by table name
            RecordIdentifier ident = event.getIdentifier();
            String table = ident.getTable();
            character fileName = new character(table.toUpperCase());
            integer fileNum = fileNums.get(fileName);
            if (fileNum == null)
            {
               Object[] args = new Object[] { fileName };
               List<integer> list = persistence.list(HQL_FILENUM,
                                                     args,
                                                     TYPES_FILENUM,
                                                     1,
                                                     0,
                                                     true);
               if (list == null)
               {
                  String msg = String.format("Table '%s' not found in metadata [%s]",
                                             table,
                                             event.toString());
                  throw new PersistenceException(msg);
               }
               
               fileNum = list.get(0);
               fileNums.put(fileName, fileNum);
            }
            
            // lock it for writing
            RecordIdentifier metaIdent = new RecordIdentifier(dmoClass.getName(), dmo.getId());
            persistence.lock(LockType.EXCLUSIVE, metaIdent, true);
            
            // set user name and user id
            String subject = event.getLocker().getSubject();
            character userName = new character(subject);
            proxy.setLockName(userName);
            proxy.setLockUsr(event.getUserid());
            
            // assign other properties that will not change with new events
            proxy.setLockId(new int64(key));    // re-use primary key
            proxy.setLockTable(fileNum);
            proxy.setLockRecid(new int64(getLockRecordID(event)));
            proxy.setLockName(new character(event.getLocker().getSubject()));
         }
         catch (InstantiationException | IllegalAccessException exc)
         {
            String msg = String.format("Error creating meta lock record for event %s",
                                       event.toString());
            
            throw new PersistenceException(msg, exc);
         }
         
         return dmo;
      }
      
      /**
       * Invocation handler which delegates calls to the {@link MinimalLock} proxy methods to the
       * backing DMO instance.
       */
      private static class Handler
      implements InvocationHandler
      {
         /** DMO delegate whose methods ultimately are invoked */
         private Object delegate = null;
         
         /**
          * Process method invocations on the {@link MinimalLock} proxy by delegating them to the
          * backing DMO instance.
          * 
          * @param   proxy
          *          Proxy on which the enclosing class invokes methods.
          * @param   method
          *          Method which was invoked.
          * @param   args
          *          Method arguments.
          */
         @Override
         public Object invoke(Object proxy, Method method, Object[] args)
         throws Throwable
         {
            Method dmoMethod = methodMap.get(method);
            
            return dmoMethod.invoke(delegate, args);
         }
         
         /**
          * Set the backing DMO delegate instance.
          * 
          * @param   delegate
          *          DMO whose methods will be invoked.
          */
         void setDelegate(Object delegate)
         {
            this.delegate = delegate;
         }
      }
   }
   
   /**
    * Interface which defines the minimally required methods of the <code>MetaLock</code>
    * interface. This interface defines the proxy used to create and manipulate metadata lock
    * DMO instances. The DMO's methods cannot be invoked directly from code here, since the
    * <code>MetaLock</code> interface resides in an application-specific package, which is not
    * known at compile time.
    */
   public interface MinimalLock
   extends Persistable
   {
      /**
       * Getter: lockId
       * 
       * @return  lockId
       */
      public int64 getLockId();
      
      /**
       * Setter: lockId
       * 
       * @param   lockId
       *          lockId
       */
      public void setLockId(NumberType lockId);
      
      /**
       * Getter: Usr
       * 
       * @return  Usr
       */
      public integer getLockUsr();
      
      /**
       * Setter: Usr
       * 
       * @param   lockUsr
       *          Usr
       */
      public void setLockUsr(NumberType lockUsr);
      
      /**
       * Getter: Name
       * 
       * @return  Name
       */
      public character getLockName();
      
      /**
       * Setter: Name
       * 
       * @param   lockName
       *          Name
       */
      public void setLockName(Text lockName);
      
      /**
       * Getter: Type
       * 
       * @return  Type
       */
      public character getLockType();
      
      /**
       * Setter: Type
       * 
       * @param   lockType
       *          Type
       */
      public void setLockType(Text lockType);
      
      /**
       * Getter: Table
       * 
       * @return  Table
       */
      public integer getLockTable();
      
      /**
       * Setter: Table
       * 
       * @param   lockTable
       *          Table
       */
      public void setLockTable(NumberType lockTable);
      
      /**
       * Getter: RECID
       * 
       * @return  RECID
       */
      public int64 getLockRecid();
      
      /**
       * Setter: RECID
       * 
       * @param   lockRecid
       *          RECID
       */
      public void setLockRecid(NumberType lockRecid);
      
      /**
       * Getter: Flags
       * 
       * @return  Flags
       */
      public character getLockFlags();
      
      /**
       * Setter: Flags
       * 
       * @param   lockFlags
       *          Flags
       */
      public void setLockFlags(Text lockFlags);
   }
}
